#include "mpint.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

void muldivd(int j)
{
  int i;
  mp_t m= MP_INITIALISER;
  if (!j) j= 30000;
  mp_digitAtPut(&m, 0, 12345);
  assert(1 == m.width);
  for (i= 0;  i < j;  ++i) mp_mul_d(&m, 13);
  for (i= 0;  i < j;  ++i) {
    mp_mul_d(&m, 17);  mp_mul_d(&m, 19);  mp_mul_d(&m, 23);
    mp_div_d(&m, 17);  mp_div_d(&m, 19);  mp_div_d(&m, 23);
    mp_div_d(&m, 13);
  }
  assert(1 == m.width);
#if !defined(NDEBUG)
  int c= mp_sub_d(&m, 12345);
  assert(mp_zerop(&m));
  assert(1 == c);
#endif
  mp_clear(&m);
}

void muldiv(int j)
{
  int i;
  mp_t m= MP_INITIALISER;
  mp_t n= MP_INITIALISER;
  mp_t r= MP_INITIALISER;
  if (!j) j= 120;
  mp_digitAtPut(&m, 0, 12345);
  mp_digitAtPut(&n, 0, 1);
  for (i= 1;  i <= j;  ++i) {
    mp_mul_d(&n, i);
    mp_mul(&m, &n);
  }
  for (i= j;  i > 0;  --i) {
    mp_div(&m, &r, &m, &n);
    assert(mp_zerop(&r));
    mp_div_d(&n, i);
  }
#if !defined(NDEBUG)
  int c= mp_sub_d(&m, 12345);
  assert(mp_zerop(&m));
  assert(1 == c);
#endif
  mp_clear(&m);
  mp_clear(&n);
  mp_clear(&r);
}

void add(int j)
{
  int i;
  mp_t m= MP_INITIALISER;
  mp_t n= MP_INITIALISER;
  mp_t o= MP_INITIALISER;
  if (!j) j= 60000;
  mp_digitAtPut(&m, 0, 12345);
  mp_digitAtPut(&n, 0, 1);
  for (i= 0;  i < j;  ++i) {
    mp_add(&m, &m);
    mp_copy(&o, &m);
    mp_add(&o, &o);
    mp_add(&o, &o);
    mp_add(&m, &o);
    mp_add(&n, &n);
    mp_copy(&o, &n);
    mp_add(&o, &o);
    mp_add(&o, &o);
    mp_add(&n, &o);
  }
  mp_div(&m, &o, &m, &n);
#if !defined(NDEBUG)
  int c= mp_sub_d(&m, 12345);
  assert(mp_zerop(&m));
  assert(1 == c);
  assert(mp_zerop(&o));
#endif
  mp_clear(&m);
  mp_clear(&n);
  mp_clear(&o);
}

void sub(int j)
{
  int i;
  mp_t m= MP_INITIALISER;
  mp_t n= MP_INITIALISER;
  mp_t o= MP_INITIALISER;
  mp_t p= MP_INITIALISER;
  if (!j) j= 10000;
  mp_digitAtPut(&m, 0, 123);
  for (i= 0;  i < j;  ++i) mp_mul_d(&m, 13);
  mp_digitAtPut(&n, 0, 1);
  while (mp_cmp(&m, &n) >= 0) mp_mul_d(&n, 1000);
  mp_copy(&p, &m);
  while (!mp_zerop(&n)) {
    int u= 0;
    while (mp_cmp(&m, &n) >= 0) { ++u;  mp_sub(&m, &n); }
    mp_mul_d(&o, 1000);
    mp_add_d(&o, u);
    mp_div_d(&n, 1000);
  }
  assert(0 == mp_cmp(&o, &p));
  mp_sub(&o, &p);
  assert(mp_zerop(&o));
  mp_clear(&m);
  mp_clear(&n);
  mp_clear(&o);
  mp_clear(&p);
}

void subf(int j)
{
  int i;
  mp_t m= MP_INITIALISER;
  mp_t n= MP_INITIALISER;
  mp_t o= MP_INITIALISER;
  mp_t h= MP_INITIALISER;
  if (!j) j= 80000;
  mp_digitAtPut(&m, 0, 13);
  mp_digitAtPut(&h, 0, 1);
  mp_shiftl(&h, 32);
  for (i= 0;  i < j;  ++i) {
    mp_mul_d(&m, 13);
    mp_add_d(mp_neg(mp_copy(&n, &m)), i);
    mp_subf_d(mp_copy(&o, &m), i);
    assert(0 == mp_cmp(&n, &o));
  }
  mp_clear(&m);
  mp_clear(&n);
  mp_clear(&o);
  mp_clear(&h);
}

void negnot(int j)
{
  int i;
  mp_t m= MP_INITIALISER;
  mp_t n= MP_INITIALISER;
  mp_t o= MP_INITIALISER;
  if (!j) j= 90000;
  mp_digitAtPut(&m, 0, 13);
  for (i= 0;  i < j;  ++i) {
    mp_mul_d(&m, 13);
    mp_neg(mp_copy(&n, &m));
    mp_not(mp_copy(&o, &m));
    mp_add_d(&o, 1);
    assert(0 == mp_cmp(&n, &o));
  }
  mp_clear(&m);
  mp_clear(&n);
  mp_clear(&o);
}

void andxor(int j)
{
  int i;
  mp_t m= MP_INITIALISER;
  mp_t n= MP_INITIALISER;
  mp_t o= MP_INITIALISER;
  if (!j) j= 140000;
  mp_digitAtPut(&m, 0, 23);
  mp_copy(&o, &m);
  for (i= 0;  i < j;  ++i) {
    mp_add(&o, &m);
    mp_copy(&n, &m);
    mp_xor(&m, &m);	// bitwise addition, without carries
    mp_and(&n, &n);	// carry out of each bit
    mp_shiftl(&n, 1);	// propagated to next bit
    mp_add(&m, &n);	// and added into sum
    assert(0 == mp_cmp(&m, &o));
  }
  mp_clear(&m);
  mp_clear(&n);
  mp_clear(&o);
}

void bits(int j)
{
  int i;
  mp_t m= MP_INITIALISER;
  mp_t n= MP_INITIALISER;
  mp_t o= MP_INITIALISER;
  mp_t p= MP_INITIALISER;
  mp_t s= MP_INITIALISER;
  mp_t t= MP_INITIALISER;
  if (!j) j= 2500000;
  for (i= 0;  i < 256;  ++i) {
    mp_digitAtPut(&o, 0, i);
    mp_shiftl(mp_copy(&p, &o), 8 *        i );  mp_or(&m, &p);
    mp_shiftl(mp_copy(&p, &o), 8 * (255 - i));  mp_or(&n, &p);
  }
  for (i= 0;  i < j;  ++i) {
    mp_shift(mp_copy(&s, &m), (j &  63) - 16);
    mp_shift(mp_copy(&t, &n), (j & 127) - 48);
    mp_and(mp_copy(&o, &s), &t);		// bits are both 1
    mp_not(mp_or(mp_copy(&p, &s), &t));		// bits are both 0
    mp_not(mp_or(&o, &p));			// bits are different
    mp_xor(mp_copy(&p, &s), &t);
    assert(0 == mp_cmp(&o, &p));
  }
  mp_clear(&m);
  mp_clear(&n);
  mp_clear(&o);
  mp_clear(&p);
  mp_clear(&s);
  mp_clear(&t);
}

#include <sys/time.h>

void test(char *name, void(*func)(int), int n)
{
  printf("%s\t", name);
  fflush(stdout);
  struct timeval start, stop;
  gettimeofday(&start, 0);
  func(n);
  gettimeofday(&stop, 0);
  if ((stop.tv_usec -= start.tv_usec) < 0) { stop.tv_usec += 1000000;  stop.tv_sec -= 1; }
  stop.tv_sec -= start.tv_sec;
  printf("%d.%03d\n", (int)stop.tv_sec, (int)stop.tv_usec / 1000);
  fflush(stdout);
}

int main(int argc, char **argv)
{
  do {
    ++argv;
    --argc;
    char *arg= argc ? *argv : 0;
#  define opt(X) if (!arg || !strcmp(arg, #X)) test(#X, X, 0)
    opt(muldivd);
    opt(muldiv);
    opt(add);
    opt(sub);
    opt(subf);
    opt(negnot);
    opt(andxor);
    opt(bits);
#  undef opt
  } while (argc > 1);
  printf("\n");
  return 0;
}
